#include "collision.h"
#include "ps2maths.h"
#include "defines.h"

#define DOT_PRODUCT(ux,uy,vx,vy) ((ux)*(vx) + (uy)*(vy))


// Perform a reasonably looking collision response.
void CollisionResponseFake(CBall * B, int B1, int B2)
{
	// Don't respond if last collision was between the same two balls. (prevent sticking balls)	
	if((B[B1].LastCollision == B[B2].BallNum) && (B[B2].LastCollision == B[B1].BallNum))return;

	// Swap velocity vectors
	float t0 = B[B1].Velocity.x;
	float t1 = B[B1].Velocity.y;

	B[B1].Velocity.x  = B[B2].Velocity.x;
	B[B1].Velocity.y  = B[B2].Velocity.y;
	B[B2].Velocity.x  = t0;
	B[B2].Velocity.y  = t1;

	// Set the last ball collided with (prevent sticking balls)	
	B[B1].LastCollision = B[B2].BallNum;
	B[B2].LastCollision = B[B1].BallNum;
	
	
}


// Perform a true to life collision response
void CollisionResponseTrue(CBall * B, int B1, int B2)
{
	// Don't respond if last collision was between the same two balls. (prevent sticking balls)	
	if((B[B1].LastCollision == B[B2].BallNum) && (B[B2].LastCollision == B[B1].BallNum))return;

	// Set the last ball collided with (prevent sticking balls)	
	B[B1].LastCollision = B[B2].BallNum;
	B[B2].LastCollision = B[B1].BallNum;
	
	// this function uses the mass of each ball along with the intial velocities to 
	// compute the resulting velocities

	// step 1: test each object against each other object and test for a collision
	// there are better ways to do this other than a double nested loop, but since
	// there are a small number of objects this is fine, also we want to somewhat model
	// if two or more balls hit simulataneously

	// compute the normal vector from a->b
	float nabx = (B[B1].Pos.x - B[B2].Pos.x);
	float naby = (B[B1].Pos.y - B[B2].Pos.y);
	float length = (float)Sqrt(nabx*nabx + naby*naby);

	// compute the response coordinate system axes
	// normalize normal vector
	nabx/=length;
	naby/=length;

	// compute the tangential vector perpendicular to normal, simply rotate vector 90
	// tangential is also normalized since it's just a rotated normal vector
	float tabx =  -naby;
	float taby =  nabx;

	// step 2: compute all the initial velocities
	// notation ball: (a,b) initial: i, final: f, n: normal direction, t: tangential direction
	float vait = DOT_PRODUCT(B[B1].Velocity.x, B[B1].Velocity.y, tabx, taby);
	float vain = DOT_PRODUCT(B[B1].Velocity.x, B[B1].Velocity.y, nabx, naby);
	float vbit = DOT_PRODUCT(B[B2].Velocity.x, B[B2].Velocity.y, tabx, taby);
	float vbin = DOT_PRODUCT(B[B2].Velocity.x, B[B2].Velocity.y, nabx, naby);

	// now we have all the initial velocities in terms of the n and t axes
	// step 3: compute final velocities after collision.

	float ma = 1.0f;
	float mb = 1.0f;
	float cof_E  = 1.0f;

	float vafn = (mb*vbin*(cof_E+1) + vain*(ma - cof_E*mb)) / (ma + mb);
	float vbfn = (ma*vain*(cof_E+1) - vbin*(ma - cof_E*mb)) / (ma + mb);
	
	// now the tangential components are the same before and after, so
	float vaft = vait;
	float vbft = vbit;

	// the velocity vectors are:
	// object a (vafn, vaft)
	// object b (vbfn, vbft)    

	// we need to 
	// translate back to the original x,y coordinate system, basically we need to 
	// compute the sum of the x components relative to the n,t axes and the sum of
	// the y components relative to the n,t axis, since n,t may both have x,y
	// components in the original x,y coordinate system
	float xfa = vafn*nabx + vaft*tabx;
	float yfa = vafn*naby + vaft*taby;
	float xfb = vbfn*nabx + vbft*tabx;
	float yfb = vbfn*naby + vbft*taby;

	// store results
	B[B1].Velocity.x = xfa;
	B[B1].Velocity.y = yfa;

	B[B2].Velocity.x = xfb;
	B[B2].Velocity.y = yfb;

	//update ball positions
	B[B1].Pos.x += B[B1].Velocity.x;
	B[B1].Pos.y += B[B1].Velocity.y;
	B[B2].Pos.x += B[B2].Velocity.x;
	B[B2].Pos.y += B[B2].Velocity.y;
}
	
// Perform collision detection
// Check all the balls against all the other balls
void Collision(CBall * B)
{
	float cxi, cxj, cyi, cyj, dist;

	for(int i=0; i<NUMBALLS - 1; ++i)
		for(int j= i+1; j<NUMBALLS; ++j)
		{
			cxi = B[i].Pos.x;
			cyi = B[i].Pos.y;
			cxj = B[j].Pos.x;
			cyj = B[j].Pos.y;
			
			// Use a collision bounding circle
			dist = (float)Sqrt((cxi-cxj)*(cxi-cxj) + (cyi-cyj)*(cyi-cyj));
			
			// 9 makes collisions look good!
			if(dist < (B[i].Halfw + B[j].Halfw - 9.0f))
			{
				CollisionResponseTrue(B, i, j);
				//CollisionResponseFake(B, i, j);
			}
		}
}
