#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include <memory.h>
#include <signal.h>
#include <time.h>
#include "PS2Defines.h"
#include "ps2maths.h"
#include "ps2matrix4x4.h"
#include "pad.h"
#include "sps2wrap.h"
#include "dma.h"
#include "texture.h"
#include "cball.h"
#include "collision.h"
#include "defines.h"


//function prototypes in this file
void BuildSprite2D(CBall * pSprite);
void InitObjects(CBall * B);

bool g_bLoop = true;

void sig_handle(int sig)
{
	g_bLoop = false;
}

int main()
{
	CBall Sprite[NUMBALLS];			//some ball classes to play with
	
	//seed the random number generator
  	srand(time(NULL));
	
	// Initialise control pad 0
	if(!pad_init(PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE))
	{
		printf("Failed to initialise control pad\n");
		pad_cleanup(PAD_0);
		exit(0);
	}	

	SPS2Manager.Initialise(4096);	// 4096 * 4K Pages = 16MB Total

	VIFStaticDMA.Initialise(1536);	// 1536 * 4K Pages = 6MB Static DMA
	VIFDynamicDMA.Initialise(256);	// 256 * 4K Pages * 2 Buffers =
									// 2MB Dynamic DMA

	// Register our signal function for every possible signal (i.e. ctrl + c)
	for(int sig=0; sig<128; sig++)
		signal(sig, sig_handle);
		
	// Set up the DMA packet to clear the screen.
	SPS2Manager.InitScreenClear(0x10, 0x10, 0x20);	
		
	// Load texture.
	CTexture Tex1;
	Tex1.LoadBitmap("smiling.bmp", true, false);
	
	// Configure the Sprites
	InitObjects(Sprite);

	while(g_bLoop)
	{
		// Fire off the previous frame's buffer
		VIFDynamicDMA.Fire();

		pad_update(PAD_0);
		
		// Process the objects
		for(int i=0; i<NUMBALLS; ++i)
		{
			Sprite[i].Move();
			Sprite[i].WallCollision();
			Sprite[i].Animate();
			Sprite[i].Rotate();
		}
		
		// Do collision detection
		Collision(Sprite);
	
		// Clear the screen
		SPS2Manager.BeginScene();
		
		// Upload the current texture to the free texture buffer
		Tex1.Upload(480);
		Tex1.Select();
		
		// Start DIRECT mode transfer
		VIFDynamicDMA.StartDirect();
		
		//build the GS packet
		for(int i=0; i<NUMBALLS; ++i)
		{
			BuildSprite2D(&Sprite[i]);
		}
		
		// End the direct mode
		VIFDynamicDMA.EndDirect();
	
		// Swap the buffers
		SPS2Manager.EndScene();
		
		// Check for exit condition
		if((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)) g_bLoop = false;
	}
	
	// Cleanup the Pads
	pad_cleanup(PAD_0);
	
	return 0;
}




void BuildSprite2D(CBall * pSprite)
{	

	// Get the rotation matrix from the rotation angle
	pSprite->World.RotationZ(pSprite->Rot);
	
	// These are the transformed coordinates of the sprite
	Vector4 TR, TL, BR, BL;

	//Add the position to the world matrix
	pSprite->World(3,0) = pSprite->Pos.x;
	pSprite->World(3,1) = pSprite->Pos.y;
	pSprite->World(3,2) = pSprite->Pos.z;	//this is the z depth
		
	// Apply World matrix to the untransformed coordinates
	TR = pSprite->World * pSprite->TR;
	TL = pSprite->World * pSprite->TL;
	BR = pSprite->World * pSprite->BR;
	BL = pSprite->World * pSprite->BL;
	
			
	// Create a textured tristrip
	// Add the GIFTag
	VIFDynamicDMA.Add128(GS_GIFTAG_BATCH(	4,							// NLOOP
	 										1,							// EOP
	 										1,							// PRE
	 										GS_PRIM(PRIM_TRI_STRIP, 	// PRIM
													PRIM_IIP_FLAT, 
													PRIM_TME_ON,
													PRIM_FGE_OFF, 
													PRIM_ABE_ON, 
													PRIM_AA1_OFF, 
													PRIM_FST_UV, 
													PRIM_CTXT_CONTEXT1, 
													PRIM_FIX_NOFIXDDA),					
	 										GIF_FLG_PACKED,				// FLG
	 										GS_BATCH_3(	GIF_REG_RGBAQ, 	// BATCH
	 													GIF_REG_UV, 
	 													GIF_REG_XYZ2)));

		// Add Vertex data
		
		//top left
		// Add the colour
    	VIFDynamicDMA.Add128(PACKED_RGBA(pSprite->r, pSprite->g, pSprite->b, pSprite->a));
    	
    	// Add UV texture coordinates
    	VIFDynamicDMA.Add128(PACKED_UV(	((pSprite->u)<<4), 
    									((pSprite->v)<<4)));
    	    	
    	// Add the position  
    	// The (1<<3) is a half-pixel offset to get the texture and sprite to line-up 	   	
    	VIFDynamicDMA.Add128(PACKED_XYZ2(	(2048 << 4) + (int)((TL.x)*16.0f) + (1<<3), 
    										(2048 << 4) + (int)((TL.y)*16.0f) + (1<<3), 
    										(int)TL.z, 
    										0)); 													

		//bottom left
		// Add the colour
    	VIFDynamicDMA.Add128(PACKED_RGBA(pSprite->r, pSprite->g, pSprite->b, pSprite->a));
    	
    	// Add UV texture coordinates
    	VIFDynamicDMA.Add128(PACKED_UV(	((pSprite->u)<<4), 
    									(((pSprite->v + pSprite->h)<<4))));
    	    	
    	// Add the position  
    	// The (1<<3) is a half-pixel offset to get the texture and sprite to line-up 	   	
    	VIFDynamicDMA.Add128(PACKED_XYZ2(	(2048 << 4) + (int)((BL.x)*16.0f) + (1<<3), 
    										(2048 << 4) + (int)((BL.y)*16.0f) + (1<<3), 
    										(int)BL.z, 
    										0)); 		
    	
    	//top right
    	// Add the colour
    	VIFDynamicDMA.Add128(PACKED_RGBA(pSprite->r, pSprite->g, pSprite->b, pSprite->a));
    	
    	// Add UV texture coordinates
    	VIFDynamicDMA.Add128(PACKED_UV(	((pSprite->u + pSprite->w)<<4), 
    									((pSprite->v)<<4)));
    	    	
    	// Add the position  
    	// The (1<<3) is a half-pixel offset to get the texture and sprite to line-up 	   	
    	VIFDynamicDMA.Add128(PACKED_XYZ2(	(2048 << 4) + (int)((TR.x)*16.0f) + (1<<3), 
    										(2048 << 4) + (int)((TR.y)*16.0f) + (1<<3), 
    										(int)TR.z, 
    										0)); 		
    	
    	// Bottom Right
    	// Add the colour
    	VIFDynamicDMA.Add128(PACKED_RGBA(pSprite->r, pSprite->g, pSprite->b, pSprite->a));
    	
    	// Add UV texture coordinates
    	VIFDynamicDMA.Add128(PACKED_UV(	((pSprite->u + pSprite->w)<<4), 
    									((pSprite->v + pSprite->h)<<4)));
    	    	
    	// Add the position  
    	// The (1<<3) is a half-pixel offset to get the texture and sprite to line-up 	   	
    	VIFDynamicDMA.Add128(PACKED_XYZ2(	(2048 << 4) + (int)((BR.x)*16.0f) + (1<<3), 
    										(2048 << 4) + (int)((BR.y)*16.0f) + (1<<3), 
    										(int)BR.z, 
    										0)); 		
    										
   									
}


// Initialise some of the object parameters.
void InitObjects(CBall * B)
{
	for(int i=0; i<NUMBALLS; ++i)
	{
		// Set the screen width and height
		B[i].ScreenW = (float)(sps2UScreenGetWidth()/2);
		B[i].ScreenH = (float)(sps2UScreenGetHeight()/2);
		
		// Choose a random position on screen
		B[i].Pos.x = (float)((rand()%((int)(B[i].ScreenW*2) - B[i].w)) - (int)B[i].ScreenW);
		B[i].Pos.y = (float)((rand()%((int)(B[i].ScreenH*2) - B[i].h)) - (int)B[i].ScreenW);
		
		// Choose a random velocity
		B[i].Velocity.x = ((float)(rand()%150))/10.0f - 7.5f;
		B[i].Velocity.y = ((float)(rand()%150))/10.0f - 7.5f;
		
		// Set the Ball number variable (used to prevent sticking balls)
		B[i].BallNum = i;
		
		// Set the z depth of the balls
		// Deepest balls drawn first
		B[i].Pos.z = (float)(i+1) * 500.0f;
		
		// Set a colour modulation
		// Try to make them different
		B[i].r = (char)(10 + rand()%118);
		B[i].g = (char)(10 + rand()%118);
		B[i].b = (char)(10 + rand()%118);
		
		// Some random rotations
		B[i].RotationSpeed = ((float)(rand()%1000))/3000.0f;
	}
}


